#
#  getSNPs.py
#  ENRAGE
#
# Get the matches and mismatches from the RDS file, and calulate the SNP thresholds uniqStartMin (Sl * readlength) and and totalRatio (Cl). 
# For each mismatch, choose the base change that occur most frequently (ie: has the highest number
# of indpendent reads)
# Threshold of Sl and Cl are from user input
# Sl = # of independent reads supporting a base change at position S 
# Cl = total # of all reads suppoting a base change at position S / # of all # reads that pass through position S
# Originally written by: Wendy Lee
# Last modified: December 14th, 2008 by Ali Mortazavi

import sys
from commoncode import *

print '%s: version 3.0' % sys.argv[0]

try:
    import psyco
    psyco.full()
except:
    print "psyco is not running"
    pass

def getMatchList(rds, achrom):
    spliceDict = {}
    newDict = {}
    try:
        newDict = rds.getReadsDict(fullChrom=True, bothEnds=True, noSense=True, chrom=achrom)
    except:
        newDict[achrom] = []
    
    length = len(newDict[achrom])
    try:
        spliceDict = rds.getSplicesDict(noSense=True, fullChrom=True, chrom=achrom, splitRead=True)
    except:
        spliceDict[achrom] = []
    
    if len(spliceDict[achrom]) > 0:
        newDict[achrom] += spliceDict[achrom]
        length += len(spliceDict[achrom]) / 2
    
    return newDict[achrom]

def checkBetween(matchList, pos_in):
    c = 0
    skip = 0
    
    for (start, stop) in matchList:
        if start <= pos_in <= stop:
            c += 1
        else:
            skip += 1
        if pos_in < start:
            skip -= 10
            if skip < 0:
                skip = 0
            break
    return (skip, c)

def getMismatchDict(theRDS, chrom_in):
    mainDict = {} 
    spDict = theRDS.getMismatches(mischrom=chrom_in)
    for (start, change_at, change_base, change_from) in spDict[chrom_in]:
        change = change_base + "-" + change_from
        try:
            (ucount, totalcount, back, uniqBaseDict, totalBaseDict) = mainDict[change_at]
            pos = str(start) + ':' + change
            try: 
                totalBaseDict[change] += 1
            except: 
                totalBaseDict[change] = 1
            if pos not in back:
                ucount += 1 # unique read count
                try:
                    uniqBaseDict[change] += 1 # dict contains total unique read counts
                except:
                    uniqBaseDict[change] = 1
                mainDict[change_at] = [ucount, totalcount + 1, back + "," + pos, uniqBaseDict, totalBaseDict]
            else: # if this is not a unique read
                mainDict[change_at] = [ucount, totalcount + 1, back, uniqBaseDict, totalBaseDict]
        except: 
            mainDict[change_at] = [1,1, str(start) + ":" + change, {change:1}, {change:1}] 

    return mainDict

#Main program 
if len(sys.argv) < 5:
    print 'usage: python2.5 %s samplerdsfile uniqStartMin totalRatioMin outfile [-enforceChr] [-cache pages]' % sys.argv[0]
    print 'where\n\tuniqStartMin = # of independent reads supporting a base change at position S'
    print '\ttotalRatioMin = total # of reads suppoting a base change at position S / total # reads that pass through position S'
    sys.exit(1)

outStr = "" 
hitfile = sys.argv[1]
uniqStartMin = float(sys.argv[2])
totalRatioMin = float(sys.argv[3])
writeLog('snp.log', sys.argv[0], "rdsfile: %s uniqStartMin: %1.2f totalRatioMin: %1.2f" % (hitfile, uniqStartMin, totalRatioMin))

outfilename = sys.argv[4]
cachePages = 0
doCache = False
if '-cache' in sys.argv:
    cachePages = int(sys.argv[sys.argv.index('-cache') + 1])
if cachePages > 0:
    doCache = True

forceChr = False
if '-enforceChr' in sys.argv:
    forceChr = True
    
hitRDS = readDataset(hitfile, verbose = True, cache=doCache)
if cachePages > 20000:
    hitRDS.setDBcache(cachePages)
readLength = hitRDS.getReadSize() 
outfile  = open(outfilename, 'w')
chromList = hitRDS.getChromosomes()

header = "#Sl\tCl\tchrom\tpos\tmatch\tuniqMis\t\ttotalMis\tchange" 
outfile.write(header + '\n')
for chrom_in in chromList:
    if forceChr:
        if chrom_in[:3] != 'chr':
            continue
    matchList = getMatchList(hitRDS, chrom_in)
    matchList.sort()
    print 'got match dict for %s ' % chrom_in
    mismatchDict = getMismatchDict(hitRDS, chrom_in)
    print 'got mismatch dict for %s ' % chrom_in
    start = 0
    misPositions = mismatchDict.keys()
    misPositions.sort()
    print header
    for position in misPositions:
        (uniqCnt, totalCount, back, uniqBaseDict, totalBaseDict) = mismatchDict[position]
        highestCount = 0
        highestBaseChange = 'N-N'
        highestTotalCount = 0
        for baseChange in uniqBaseDict:
            if totalBaseDict[baseChange] > highestTotalCount:
                highestBaseChange = baseChange
                highestCount = uniqBaseDict[baseChange]
                highestTotalCount = totalBaseDict[baseChange]
        Cl = 0.
        if highestCount >= uniqStartMin:
            (skip, matchCount) = checkBetween(matchList[start:], position) 
            matchCount -= totalCount
            start += skip
            if matchCount < 0:
                matchCount = 0
            Sl = highestCount/float(readLength)
            Cl = highestTotalCount/float(highestTotalCount + matchCount)
            if Cl >= totalRatioMin:
                outline = "%1.2f\t%1.2f\t%s\t%d\t%d\t%d\t\t%d\t%s" % (Sl, Cl, chrom_in, position, matchCount, highestCount, highestTotalCount, highestBaseChange)
                print outline
                outfile.write(outline + "\n") 
outfile.close()

writeLog('snp.log', sys.argv[0], "outputfile: " + outfile + "\n")

